#
# HC 0.2 (8 Feburary 2001)
#
# http://www.newdayrising.com/~laurie/python/hc/
#
# Copyright (c) 1999 - 2001, Laurence Tratt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# Neither the name of Laurence Tratt nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission. 
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE. 
#






import htmlentitydefs, string, types, UserList






#
# Module constants
#

_NEWLINE = "\n"
_TAB = "\t"






class _Base(UserList.UserList):


	start_tag = 1
	end_tag = 1

	external_lines__beginning = 0
	external_lines__end = 0

	internal_lines__beginning = 0
	internal_lines__end = 0

	contents__indentable = 1
	contents__processable = 1

	contents__indent = 0



	def __init__(self, object = None, *attributes):

		self.data = []
		self._attributes = {}

		if object is not None: self.append(object)
		for attribute in attributes:
			self._attributes[attribute[0]] = attribute[1]



	def __repr__(self):

		return "<HC.%s [%s]>" % (self.__class__, string.join(map(repr, self.data)))



	def add_attribute(self, attribute):

		self._attributes[attribute[0]] = attribute[1]



	def get_attributes(self):

		return self._attributes



	def html(self):

		return _output(self)





class _Blank(_Base):

	start_tag = 0
	end_tag = 0





class A(_Base):


	tag = "A"


	def __init__(self, object = None, href = None, name = None, *attributes):

		apply(_Base.__init__, [self, object] + list(attributes))

		if href: self._attributes["HRef"] = "\"" + href + "\""
		if name: self._attributes["Name"] = "\"" + name + "\""





class Address(_Base):


	tag = "Address"


	def __init__(self, object = None, align = None, *attributes):

		apply(_Base.__init__, [self, object] + list(attributes))

		if align: self._attributes["Align"] = align





class Big(_Base):


	tag = "Big"





class Body(_Base):


	tag = "Body"

	internal_lines__beginning = 1
	internal_lines__end = 1

	external_lines__beginning = 3

	contents__indent = 1





class Br(_Base):


	tag = "Br"

	end_tag = 0

	external_lines__end = 1





class Code(_Base):


	tag = "Code"





class DD(_Base):


	tag = "DD"

	external_lines__beginning = 1
	external_lines__end = 1

	contents__indent = 1






class DL(_Base):


	tag = "DL"

	external_lines__beginning = 1
	external_lines__end = 1

	contents__indent = 1





class DT(_Base):


	tag = "DT"





class Em(_Base):


	tag = "Em"





class Font(_Base):


	tag = "Font"


	def __init__(self, object = None, size = None, *attributes):

		apply(_Base.__init__, [self, object] + list(attributes))

		if size is not None: self._attributes["Size"] = "\"%s\"" % `size`





class Form(_Base):


	tag = "Form"


	def __init__(self, object = None, action = None, method = None, *attributes):

		apply(_Base.__init__, [self, object] + list(attributes))

		if action is not None: self._attributes["Action"] = "\"%s\"" % action
		if method is not None: self._attributes["Method"] = method





class Head(_Base):


	tag = "Head"


	external_lines__end = 2

	internal_lines__beginning = 1
	internal_lines__end = 1

	contents__indent = 1




class HTML(_Base):


	tag = "HTML"

	internal_lines__beginning = 1
	internal_lines__end = 1

	contents__indent = 0





class _Hx(_Base):


	external_lines__beginning = 3
	external_lines__end = 2



	def __init__(self, item = None, align = None):

		_Base.__init__(self)

		if item is not None: self.append(item)
		if align is not None: self.add_attribute(("Align", align))





class H1(_Hx):


	tag = "H1"





class H2(_Hx):


	tag = "H2"





class H3(_Hx):


	tag = "H3"





class H4(_Hx):


	tag = "H4"





class H5(_Hx):


	tag = "H5"





class H6(_Hx):


	tag = "H6"





class HR(_Base):


	tag = "HR"
	
	end_tag = 0


	external_lines__beginning = 2
	external_lines__after = 2





class Img(_Base):


	tag = "Img"


	def __init__(self, object = None, src = None, alt = None, align = None, border = None, *attributes):

		_Base.__init__(object, attributes)

		if src is not None: self._attributes["Src"] = "\"%s\"" % src
		if alt is not None: self._attributes["Alt"] = "\"%s\"" % alt
		if align is not None: self._attributes["Align"] = align
		if border is not None: self._attributes["Border"] = `border`





class Input(_Base):


	tag = "Input"


	def __init__(self, object = None, name = None, type = None, value = None, checked = None, maxlength = None, *attributes):

		_Base.__init__(object, attributes)

		if name is not None: self._attributes["Name"] = "\"%s\"" % name
		if type is not None: self._attributes["Type"] = type
		if checked: self._attributes["Checked"] = ""
		if value is not None: self._attributes["Value"] = "\"%s\"" % value
		if maxlength is not None: self._attributes["MaxLength"] = `maxlength`





class LI(_Base):


	tag = "LI"

	contents__indent = 1
	
	internal_lines__beginning = 1
	internal_lines__end = 1
	
	external_lines__beginning = 1
	external_lines__end = 1





class LH(_Base):


	tag = "LH"


	external_lines__beginning = 1
	external_lines__end = 1





class Meta(_Base):


	tag = "Meta"


	def __init__(self, object = None, name = None, content = None, http_equiv = None, url = None, *attributes):

		_Base.__init__(object, attributes)

		if object is not None: self._attributes["Object"] = "\"%s\"" % object
		if name is not None: self._attributes["name"] = "\"%s\"" % name
		if http_equiv is not None: self._attributes["HTTP-Equiv"] = "\"%s\"" % http_equiv
		if url is not None: self._attributes["url"] = "\"%s\"" % url




class Option(_Base):


	tag = "Option"


	def __init__(self, object = None, value = None, selected = None, *attributes):

		_Base.__init__(object, attributes)

		if value is not None: self._attributes["Value"] = "\"%s\"" % value
		if selected: self._attributes["Selected"] = ""





class P(_Base):


	tag = "P"

	external_lines__beginning = 2
	external_lines__end = 1


	def __init__(self, object = None, align = None, *attributes):

		apply(_Base.__init__, [self, object] + list(attributes))

		if align: self._attributes["Align"] = align





class Pre(_Base):


	tag = "Pre"

	contents__indentable = 0





class Select(_Base):


	tag = "Select"

	external_lines__beginning = 1
	external_lines__end = 1

	internal_lines__beginning = 1
	internal_lines__end = 1

	contents__indent = 1


	def __init__(self, object = None, name = None, multiple = None, *attributes):

		_Base.__init__(object, attributes)

		if name is not None: self._attributes["Name"] = "\"%s\"" % name
		if multiple: self._attributes["Multiple"] = ""





class Strong(_Base):


	tag = "Strong"





class Sup(_Base):


	tag = "Sup"





class Table(_Base):


	tag = "Table"

	external_lines__beginning = 2
	external_lines__end = 2

	internal_lines__beginning = 1
	internal_lines__end = 1

	contents__indent = 1


	def __init__(self, object = None, align = None, border = None, cellspacing = None, cellpadding = None, *attributes):

		apply(_Base.__init__, [self, object] + list(attributes))

		if align is not None: self._attributes["Align"] = align
		if border is not None: self._attributes["Border"] = `border`
		if cellspacing is not None:
			if type(cellspacing) == types.StringType:
				self._attributes["CellSpacing"] = "\"%s\"" % cellspacing
			else:
				self._attributes["CellSpacing"] = "\"%s\"" % `cellspacing`
		if cellpadding is not None:
			if type(cellspacing) == types.StringType:
				self._attributes["CellPadding"] = "\"%s\"" % cellpadding
			else:
				self._attributes["CellPadding"] = "\"%s\"" % `cellpadding`





class TD(_Base):


	tag = "TD"

	external_lines__beginning = 1
	external_lines__end = 1

	internal_lines__beginning = 1
	internal_lines__end = 1

	contents__indent = 1


	def __init__(self, object = None, align = None, valign = None, colspan = None, rowspan = None, width = None, nowrap = None, *attributes):

		apply(_Base.__init__, [self, object] + list(attributes))

		if align is not None: self._attributes["Align"] = align
		if valign is not None: self._attributes["VAlign"] = valign
		if colspan is not None: self._attributes["ColSpan"] = `colspan`
		if rowspan is not None: self._attributes["RowSpan"] = `rowspan`
		if width is not None:
			if type(width) == types.StringType:
				self._attributes["Width"] = "\"%s\"" % width
			else:
				self._attributes["Width"] = "\"%s\"" % `width`
		if nowrap: self._attributes["NoWrap"] = ""





class TH(_Base):


	tag = "TH"


	external_lines__beginning = 1
	external_lines__end = 1

	internal_lines__beginning = 1
	internal_lines__end = 1

	contents__indent = 1


	def __init__(self, object = None, align = None, valign = None, *attributes):

		apply(_Base.__init__, [self, object] + list(attributes))
		
		if align is not None: self._attributes["Align"] = align
		if valign is not None: self._attributes["VAlign"] = valign





class TR(_Base):


	tag = "TR"

	external_lines__beginning = 1
	external_lines__end = 1

	internal_lines__beginning = 1
	internal_lines__end = 1

	contents__indent = 1


	def __init__(self, object = None, align = None, valign = None, *attributes):

		apply(_Base.__init__, [self, object] + list(attributes))
		
		if align is not None: self._attributes["Align"] = align
		if valign is not None: self._attributes["VAlign"] = valign





class Title(_Base):


	tag = "Title"





class TT(_Base):


	tag = "TT"





class U(_Base):


	tag = "U"





class UL(_Base):


	tag = "UL"

	external_lines__beginning = 1
	external_lines__end = 1

	internal_lines__beginning = 1
	internal_lines__end = 1

	contents__indent = 1






#
#
#

def _output(object):

	# This method is fairly complicated, so deserves explanation
	# Because it is also *very* slow, it has been optimised. This doesn't necessarily help
	#   readibility
	#
	# To output human friendly HTML, we take a two phase approach:
	#
	#   1  Flatten the recursive object structure into a single list
	#   2  Convert the flattened list into text
	#
	# We need this two phase approach because the second phase needs to know about information
	#   forward of the current position in the list and so the list must be created before hand
	#
	# The format of the list is simple; there are only two possible types of element in it:
	#
	#   1  "Tag tuples"
	#       These are tuples of the format:
	#         (1, <object>)        for a *start* tag
	#         (0, <object>)        for a *end* tag
	#       The <object> is a standard HC object; this means we can maintain all the necessary
	#         information about this particular HTML object without having to copy it temporarily;
	#         we *only* use the public information about the HTML object and are not at all
	#         interested in the user data it may or may not hold
	#   2  Strings

	from types import * # Optimisation

	# Phase 1  :  Flatten the recursive object structure in 'object' into a single list 'items'

	items = []

	# We keep a stack 'object_stack' which contains length 3 tuples:
	#   (<object>, <data>, <position in data>)
	# Although <data> initially contains the same data as <object>, it *must* be a different
	#   physical copy because we allow ourselves to alter <data> if it contains sub-lists, and we
	#   don't want to alter the users objects

#	object_stack = stack__create([[object, object[ : ], 0]])
	object_stack = [[object, object[ : ], 0]]
	while len(object_stack) > 0:
		object, object__objects, object__objects_pos = object_stack[-1]
		if object__objects_pos == 0:
			items.append((1, object))

		# This while loop will flatten sub-lists into the main list
		# So if the current item in the list is [1, 2, [3, 4]] the overall list will be altered so
		#   the single list element is replaced with 1, 2, 3, 4 (in this example, the list will thus
		#   be lengthened)
		# It is possible that the sub-list that we come across could be empty, and so if the main
		#   list itself contains no other items, then the whole thing ends up as a blank list (eg
		#   if the main list is [[], [[]], []] it will flatten to []

		while len(object__objects) > object__objects_pos:
			item = object__objects[object__objects_pos]

			if type(item) == ListType or isinstance(item, _Blank):
				# We've found a sublist, so flatten it into the main list

				object__objects[object__objects_pos : object__objects_pos + 1] = item
			else:
				break
		else:
			# If we've got here, then there's nothing in this list to process

			items.append((0, object))
			object_stack.pop()
			continue

		# Move the list position on 1 for the current object *before* we possibly add an object to
		#   the stack. This is a little ugly, but convenient

		object_stack[-1][2] = object_stack[-1][2] + 1

		if type(item) == StringType:
			items.append(item)
		else:
			object_stack.append([item, item[ : ], 0])

	#print string.join(map(str, items), "\n\n")

	# Phase 2 : Convert the flattened list in 'items' into a string.join'able list 'output'

	items__pos = 0

	# To convert the list in 'items'

	tag_stack = []
	contents__indentable = contents__processable = 1
	contents__indent = 0
	output = []
	new_text_needs_to_be_indented = 1
	while items__pos < len(items):
		item = items[items__pos]
		if type(item) == TupleType:
			object = item[1]
			if item[0] == 1:
				if object.start_tag:
					if items__pos > 0 and not (type(items[items__pos - 1]) == TupleType and items[items__pos - 1][0] == 1 and contents__indent != tag_stack[-1][2]):
						if type(previous_item) == StringType:
							if object.external_lines__beginning > 0:
								output.append(_NEWLINE * object.external_lines__beginning)
								new_text_needs_to_be_indented = 1
						else:
							previous_object = previous_item[1]
							if previous_item[0] == 1 and object.external_lines__beginning - previous_object.internal_lines__beginning > 0:
								output.append(_NEWLINE * (object.external_lines__beginning - previous_object.internal_lines__beginning))
								new_text_needs_to_be_indented = 1
							elif previous_item[0] == 0 and object.external_lines__beginning - previous_object.external_lines__end > 0:
								output.append(_NEWLINE * (object.external_lines__beginning - previous_object.external_lines__end))
								new_text_needs_to_be_indented = 1
					if contents__indentable and new_text_needs_to_be_indented:
						output.append(_TAB * contents__indent)
					output.append("<")
					output.append(object.tag)
					if len(object.get_attributes()) > 0:
						for attribute in object.get_attributes().items():
							output.append(" ")
							if attribute[1] == "":
								output.append(attribute[0])
							else:
								output.append("%s = %s" % (attribute[0], attribute[1]))
					output.append(">")
					new_text_needs_to_be_indented = 0
					if items__pos < len(items) and object.internal_lines__beginning > 0:
						output.append(_NEWLINE * object.internal_lines__beginning)
						new_text_needs_to_be_indented = 1
				tag_stack.append((contents__indentable, contents__processable, contents__indent))
				if not object.contents__indentable:
					contents__indentable = 0
					contents__indent = 0
				else:
					contents__indent = contents__indent + object.contents__indent
				if not object.contents__processable:
					contents__processable = 0
			else:
				if object.end_tag:
					if type(previous_item) == StringType:
						if object.internal_lines__end > 0:
							output.append(_NEWLINE * object.internal_lines__end)
							new_text_needs_to_be_indented = 1
					else:
						previous_object = previous_item[1]
						if previous_item[0] == 1 and object.internal_lines__end - previous_object.internal_lines__beginning > 0:
							output.append(_NEWLINE * (object.internal_lines__end - previous_object.internal_lines__beginning))
							new_text_needs_to_be_indented = 1
						elif previous_item[0] == 0 and object.internal_lines__end - previous_object.external_lines__end > 0:
							output.append(_NEWLINE * (object.internal_lines__end - previous_object.external_lines__end))
							new_text_needs_to_be_indented = 1
					if contents__indentable and new_text_needs_to_be_indented:
						output.append(_TAB * tag_stack[-1][2])
					output.append("</%s>" % object.tag)
					new_text_needs_to_be_indented = 0
					if items__pos < len(items) and object.external_lines__end > 0:
						if items__pos + 1 < len(items) and type(items[items__pos + 1]) == TupleType and items[items__pos + 1][0] == 0 and contents__indent != tag_stack[-2][2]:
							output.append(_NEWLINE)
							new_text_needs_to_be_indented = 1
						else:
							output.append(_NEWLINE * object.external_lines__end)
							new_text_needs_to_be_indented = 1
				contents__indentable, contents__processable, contents__indent = tag_stack.pop()
		else:
			if contents__indentable and new_text_needs_to_be_indented:
				output.append(_TAB * contents__indent)
			if contents__processable:
				item = _process(item)
			if contents__indentable:
				output.append(string.replace(item, _NEWLINE, "%s%s" % (_NEWLINE, _TAB * contents__indent)))
			else:
				output.append(item)
			if len(item) > 0 and item[-1] != _NEWLINE:
				new_text_needs_to_be_indented = 0
		previous_item = item
		items__pos = items__pos + 1

	return string.join(output, "")



HTML_ENTITIES = htmlentitydefs.entitydefs
del HTML_ENTITIES["amp"]
HTML_ENTITIES = HTML_ENTITIES.items()

def _process(text):

	# In processing normal characters into valid HTML entities, we have to make a special case of the
	#   '&' character, because HTML entities are of the form '&<entity>;'

	text = text.replace("&", "&amp;")
	for replacement, char in HTML_ENTITIES:
		text = text.replace(char, "&" + replacement + ";")

	return text



#
# A simple stack
# This allows HC.py to be a completely self-contained module
#

def stack__create(stack = None):

	return _Stack(stack)



class _Stack:


	def __init__(self, stack = None):

		if stack:
			self.stack = stack
		else:
			self.stack = []



	def __repr__(self):

		return "<stack " + `self.stack` + ">"



	def __len__(self):

		return len(self.stack)



	def peek(self, offset = 1):

		return self.stack[len(self.stack) - offset]



	def pop(self):

		item = self.stack[len(self.stack) - 1]
		del(self.stack[len(self.stack) - 1])

		return item



	def push(self, item):

		self.stack.append(item)




if __name__ == "__main__":
	page = HTML()
	page.append(Head(Title()))
	body = Body()
	h1 = H1(["a", ["b", ["c"]], []])
#	h1.add_attribute(("Align", "Center"))
#	h1.append("woo")
	body.append("A woo to you\n& you")
	body.append(h1)
	page.append(body)
	print page.html()
